@@ -130,7 +130,7 @@ Request request_path(const std::string_view &uri) {
     req.path = std::string(uri.data() + u.field_data[URLPARSE_PATH].off,
                            u.field_data[URLPARSE_PATH].len);
     if (req.path.find('%') != std::string::npos) {
-      req.path = util::percent_decode(std::begin(req.path), std::end(req.path));
+      req.path = util::percent_decode(req.path);
     }
     if (!req.path.empty() && req.path.back() == '/') {
       req.path += "index.html";
@@ -165,7 +165,7 @@ std::unordered_map<std::string, FileEntry> file_cache;
 
 std::pair<FileEntry, int> Stream::open_file(const std::string &path) {
   auto it = file_cache.find(path);
-  if (it != std::end(file_cache)) {
+  if (it != std::ranges::end(file_cache)) {
     return {(*it).second, 0};
   }
 
@@ -348,6 +348,9 @@ Handler::Handler(struct ev_loop *loop, Server *server)
       true
 #endif // !defined(UDP_SEGMENT)
     },
+    close_wait_{
+      .next_pkts_recv = 1,
+    },
     tx_{
       .data = std::unique_ptr<uint8_t[]>(new uint8_t[64_k]),
     } {
@@ -488,7 +491,7 @@ int acked_stream_data_offset(ngtcp2_conn *conn, int64_t stream_id,
 int Handler::acked_stream_data_offset(int64_t stream_id, uint64_t offset,
                                       uint64_t datalen) {
   auto it = streams_.find(stream_id);
-  assert(it != std::end(streams_));
+  assert(it != std::ranges::end(streams_));
   auto &stream = (*it).second;
   (void)stream;
 
@@ -512,7 +515,7 @@ void Handler::on_stream_open(int64_t stream_id) {
   }
   auto it = streams_.find(stream_id);
   (void)it;
-  assert(it == std::end(streams_));
+  assert(it == std::ranges::end(streams_));
   streams_.emplace(stream_id, std::make_unique<Stream>(stream_id, this));
 }
 
@@ -530,8 +533,11 @@ int stream_close(ngtcp2_conn *conn, uint32_t flags, int64_t stream_id,
 
 namespace {
 void rand(uint8_t *dest, size_t destlen, const ngtcp2_rand_ctx *rand_ctx) {
-  auto dis = std::uniform_int_distribution<uint8_t>();
-  std::generate(dest, dest + destlen, [&dis]() { return dis(randgen); });
+  auto rv = util::generate_secure_random({dest, destlen});
+  if (rv != 0) {
+    assert(0);
+    abort();
+  }
 }
 } // namespace
 
@@ -639,7 +645,7 @@ int extend_max_stream_data(ngtcp2_conn *conn, int64_t stream_id,
 
 int Handler::extend_max_stream_data(int64_t stream_id, uint64_t max_data) {
   auto it = streams_.find(stream_id);
-  assert(it != std::end(streams_));
+  assert(it != std::ranges::end(streams_));
   auto &stream = (*it).second;
 
   if (nghttp3_buf_len(&stream->respbuf)) {
@@ -668,45 +674,26 @@ int Handler::init(const Endpoint &ep, const Address &local_addr,
                   std::span<const uint8_t> token, ngtcp2_token_type token_type,
                   uint32_t version, TLSServerContext &tls_ctx) {
   auto callbacks = ngtcp2_callbacks{
-    nullptr, // client_initial
-    ngtcp2_crypto_recv_client_initial_cb,
-    ::recv_crypto_data,
-    ::handshake_completed,
-    nullptr, // recv_version_negotiation
-    ngtcp2_crypto_encrypt_cb,
-    ngtcp2_crypto_decrypt_cb,
-    do_hp_mask,
-    ::recv_stream_data,
-    ::acked_stream_data_offset,
-    stream_open,
-    stream_close,
-    nullptr, // recv_stateless_reset
-    nullptr, // recv_retry
-    nullptr, // extend_max_local_streams_bidi
-    nullptr, // extend_max_local_streams_uni
-    rand,
-    get_new_connection_id,
-    remove_connection_id,
-    ::update_key,
-    path_validation,
-    nullptr, // select_preferred_addr
-    nullptr, // stream_reset
-    nullptr, // extend_max_remote_streams_bidi
-    nullptr, // extend_max_remote_streams_uni
-    ::extend_max_stream_data,
-    nullptr, // dcid_status
-    nullptr, // handshake_confirmed
-    nullptr, // recv_new_token
-    ngtcp2_crypto_delete_crypto_aead_ctx_cb,
-    ngtcp2_crypto_delete_crypto_cipher_ctx_cb,
-    nullptr, // recv_datagram
-    nullptr, // ack_datagram
-    nullptr, // lost_datagram
-    ngtcp2_crypto_get_path_challenge_data_cb,
-    nullptr, // stream_stop_sending
-    ngtcp2_crypto_version_negotiation_cb,
-    nullptr, // recv_rx_key
-    nullptr, // recv_tx_key
+    .recv_client_initial = ngtcp2_crypto_recv_client_initial_cb,
+    .recv_crypto_data = ::recv_crypto_data,
+    .handshake_completed = ::handshake_completed,
+    .encrypt = ngtcp2_crypto_encrypt_cb,
+    .decrypt = ngtcp2_crypto_decrypt_cb,
+    .hp_mask = do_hp_mask,
+    .recv_stream_data = ::recv_stream_data,
+    .acked_stream_data_offset = ::acked_stream_data_offset,
+    .stream_open = stream_open,
+    .stream_close = stream_close,
+    .rand = rand,
+    .get_new_connection_id = get_new_connection_id,
+    .remove_connection_id = remove_connection_id,
+    .update_key = ::update_key,
+    .path_validation = path_validation,
+    .extend_max_stream_data = ::extend_max_stream_data,
+    .delete_crypto_aead_ctx = ngtcp2_crypto_delete_crypto_aead_ctx_cb,
+    .delete_crypto_cipher_ctx = ngtcp2_crypto_delete_crypto_cipher_ctx_cb,
+    .get_path_challenge_data = ngtcp2_crypto_get_path_challenge_data_cb,
+    .version_negotiation = ngtcp2_crypto_version_negotiation_cb,
   };
 
   scid_.datalen = NGTCP2_SV_SCIDLEN;
@@ -766,8 +753,8 @@ int Handler::init(const Endpoint &ep, const Address &local_addr,
     settings.pmtud_probeslen = config.pmtud_probes.size();
 
     if (!config.max_udp_payload_size) {
-      settings.max_tx_udp_payload_size = *std::max_element(
-        std::begin(config.pmtud_probes), std::end(config.pmtud_probes));
+      settings.max_tx_udp_payload_size =
+        *std::ranges::max_element(config.pmtud_probes);
     }
   }
 
@@ -832,15 +819,17 @@ int Handler::init(const Endpoint &ep, const Address &local_addr,
   }
 
   auto path = ngtcp2_path{
-    {
-      const_cast<sockaddr *>(&local_addr.su.sa),
-      local_addr.len,
-    },
-    {
-      const_cast<sockaddr *>(sa),
-      salen,
-    },
-    const_cast<Endpoint *>(&ep),
+    .local =
+      {
+        .addr = const_cast<sockaddr *>(&local_addr.su.sa),
+        .addrlen = local_addr.len,
+      },
+    .remote =
+      {
+        .addr = const_cast<sockaddr *>(sa),
+        .addrlen = salen,
+      },
+    .user_data = const_cast<Endpoint *>(&ep),
   };
   if (auto rv =
         ngtcp2_conn_server_new(&conn_, dcid, &scid_, &path, version, &callbacks,
@@ -868,15 +857,17 @@ int Handler::feed_data(const Endpoint &ep, const Address &local_addr,
                        const ngtcp2_pkt_info *pi,
                        std::span<const uint8_t> data) {
   auto path = ngtcp2_path{
-    {
-      const_cast<sockaddr *>(&local_addr.su.sa),
-      local_addr.len,
-    },
-    {
-      const_cast<sockaddr *>(sa),
-      salen,
-    },
-    const_cast<Endpoint *>(&ep),
+    .local =
+      {
+        .addr = const_cast<sockaddr *>(&local_addr.su.sa),
+        .addrlen = local_addr.len,
+      },
+    .remote =
+      {
+        .addr = const_cast<sockaddr *>(sa),
+        .addrlen = salen,
+      },
+    .user_data = const_cast<Endpoint *>(&ep),
   };
 
   if (auto rv = ngtcp2_conn_read_pkt(conn_, &path, pi, data.data(), data.size(),
@@ -973,6 +964,8 @@ int Handler::write_streams() {
     std::span{tx_.data.get(), std::max(ngtcp2_conn_get_send_quantum(conn_),
                                        path_max_udp_payload_size)};
   auto buf = txbuf;
+  auto pkt = std::span<const uint8_t>{};
+  auto extra_pkt = std::span<const uint8_t>{};
 
   ngtcp2_path_storage_zero(&ps);
   ngtcp2_path_storage_zero(&prev_ps);
@@ -980,11 +973,12 @@ int Handler::write_streams() {
   for (;;) {
     int64_t stream_id = -1;
     size_t vcnt = 0;
-    uint32_t flags = NGTCP2_WRITE_STREAM_FLAG_MORE;
+    uint32_t flags =
+      NGTCP2_WRITE_STREAM_FLAG_MORE | NGTCP2_WRITE_STREAM_FLAG_PADDING;
     Stream *stream = nullptr;
 
     if (!sendq_.empty() && ngtcp2_conn_get_max_data_left(conn_)) {
-      stream = *std::begin(sendq_);
+      stream = *std::ranges::begin(sendq_);
 
       stream_id = stream->stream_id;
       vec.base = stream->respbuf.pos;
@@ -1006,13 +1000,13 @@ int Handler::write_streams() {
       case NGTCP2_ERR_STREAM_DATA_BLOCKED:
       case NGTCP2_ERR_STREAM_SHUT_WR:
         assert(ndatalen == -1);
-        sendq_.erase(std::begin(sendq_));
+        sendq_.erase(std::ranges::begin(sendq_));
         continue;
       case NGTCP2_ERR_WRITE_MORE:
         assert(ndatalen >= 0);
         stream->respbuf.pos += ndatalen;
         if (nghttp3_buf_len(&stream->respbuf) == 0) {
-          sendq_.erase(std::begin(sendq_));
+          sendq_.erase(std::ranges::begin(sendq_));
         }
         continue;
       }
@@ -1026,38 +1020,24 @@ int Handler::write_streams() {
     } else if (ndatalen >= 0) {
       stream->respbuf.pos += ndatalen;
       if (nghttp3_buf_len(&stream->respbuf) == 0) {
-        sendq_.erase(std::begin(sendq_));
+        sendq_.erase(std::ranges::begin(sendq_));
       }
     }
 
     if (nwrite == 0) {
-      auto data = std::span{std::begin(txbuf), std::begin(buf)};
-      if (!data.empty()) {
-        auto &ep = *static_cast<Endpoint *>(prev_ps.path.user_data);
-
-        if (auto [rest, rv] = server_->send_packet(
-              ep, no_gso_, prev_ps.path.local, prev_ps.path.remote, prev_ecn,
-              data, gso_size);
-            rv != NETWORK_ERR_OK) {
-          assert(NETWORK_ERR_SEND_BLOCKED == rv);
-
-          on_send_blocked(ep, prev_ps.path.local, prev_ps.path.remote, prev_ecn,
-                          rest, gso_size);
-
-          start_wev_endpoint(ep);
-        }
+      pkt = std::span{std::ranges::begin(txbuf), std::ranges::begin(buf)};
+      if (pkt.empty()) {
+        return 0;
       }
 
-      // We are congestion limited.
-      ngtcp2_conn_update_pkt_tx_time(conn_, ts);
-      return 0;
+      break;
     }
 
-    auto last_pkt = std::begin(buf);
+    auto last_pkt_pos = std::ranges::begin(buf);
 
     buf = buf.subspan(nwrite);
 
-    if (last_pkt == std::begin(txbuf)) {
+    if (last_pkt_pos == std::ranges::begin(txbuf)) {
       ngtcp2_path_copy(&prev_ps.path, &ps.path);
       prev_ecn = pi.ecn;
       gso_size = nwrite;
@@ -1065,70 +1045,52 @@ int Handler::write_streams() {
                static_cast<size_t>(nwrite) > gso_size ||
                (gso_size > path_max_udp_payload_size &&
                 static_cast<size_t>(nwrite) != gso_size)) {
-      auto &ep = *static_cast<Endpoint *>(prev_ps.path.user_data);
-      auto data = std::span{std::begin(txbuf), last_pkt};
-
-      if (auto [rest, rv] =
-            server_->send_packet(ep, no_gso_, prev_ps.path.local,
-                                 prev_ps.path.remote, prev_ecn, data, gso_size);
-          rv != 0) {
-        assert(NETWORK_ERR_SEND_BLOCKED == rv);
-
-        on_send_blocked(ep, prev_ps.path.local, prev_ps.path.remote, prev_ecn,
-                        rest, gso_size);
-
-        data = std::span{last_pkt, std::begin(buf)};
-        on_send_blocked(*static_cast<Endpoint *>(ps.path.user_data),
-                        ps.path.local, ps.path.remote, pi.ecn, data,
-                        data.size());
-
-        start_wev_endpoint(ep);
-      } else {
-        auto &ep = *static_cast<Endpoint *>(ps.path.user_data);
-        auto data = std::span{last_pkt, std::begin(buf)};
-
-        if (auto [rest, rv] =
-              server_->send_packet(ep, no_gso_, ps.path.local, ps.path.remote,
-                                   pi.ecn, data, data.size());
-            rv != 0) {
-          assert(rest.size() == data.size());
-          assert(NETWORK_ERR_SEND_BLOCKED == rv);
-
-          on_send_blocked(ep, ps.path.local, ps.path.remote, pi.ecn, rest,
-                          rest.size());
-
-          start_wev_endpoint(ep);
-        }
-      }
-
-      ngtcp2_conn_update_pkt_tx_time(conn_, ts);
-      return 0;
+      pkt = std::span{std::ranges::begin(txbuf), last_pkt_pos};
+      extra_pkt = std::span{last_pkt_pos, std::ranges::begin(buf)};
+      break;
     }
 
     if (buf.size() < path_max_udp_payload_size ||
         static_cast<size_t>(nwrite) < gso_size) {
-      auto &ep = *static_cast<Endpoint *>(ps.path.user_data);
-      auto data = std::span{std::begin(txbuf), std::begin(buf)};
-
-      if (auto [rest, rv] = server_->send_packet(
-            ep, no_gso_, ps.path.local, ps.path.remote, pi.ecn, data, gso_size);
-          rv != 0) {
-        assert(NETWORK_ERR_SEND_BLOCKED == rv);
-
-        on_send_blocked(ep, ps.path.local, ps.path.remote, pi.ecn, rest,
-                        gso_size);
-
-        start_wev_endpoint(ep);
-      }
-
-      ngtcp2_conn_update_pkt_tx_time(conn_, ts);
-      return 0;
+      pkt = std::span{std::ranges::begin(txbuf), std::ranges::begin(buf)};
+      break;
     }
   }
+
+  assert(!pkt.empty());
+
+  if (send_packet(prev_ps.path, prev_ecn, pkt, gso_size) != 0) {
+    if (!extra_pkt.empty()) {
+      on_send_blocked(ps.path, pi.ecn, extra_pkt, extra_pkt.size());
+    }
+  } else if (!extra_pkt.empty()) {
+    send_packet(ps.path, pi.ecn, extra_pkt, extra_pkt.size());
+  }
+
+  ngtcp2_conn_update_pkt_tx_time(conn_, ts);
+
+  return 0;
 }
 
-void Handler::on_send_blocked(Endpoint &ep, const ngtcp2_addr &local_addr,
-                              const ngtcp2_addr &remote_addr, unsigned int ecn,
+int Handler::send_packet(const ngtcp2_path &path, unsigned int ecn,
+                         std::span<const uint8_t> data, size_t gso_size) {
+  auto &ep = *static_cast<Endpoint *>(path.user_data);
+  auto [rest, rv] = server_->send_packet(ep, no_gso_, path.local, path.remote,
+                                         ecn, data, gso_size);
+  if (rv != 0) {
+    assert(NETWORK_ERR_SEND_BLOCKED == rv);
+
+    on_send_blocked(path, ecn, rest, gso_size);
+
+    start_wev_endpoint(ep);
+
+    return rv;
+  }
+
+  return 0;
+}
+
+void Handler::on_send_blocked(const ngtcp2_path &path, unsigned int ecn,
                               std::span<const uint8_t> data, size_t gso_size) {
   assert(tx_.num_blocked || !tx_.send_blocked);
   assert(tx_.num_blocked < 2);
@@ -1138,12 +1100,12 @@ void Handler::on_send_blocked(Endpoint &ep, const ngtcp2_addr &local_addr,
 
   auto &p = tx_.blocked[tx_.num_blocked++];
 
-  memcpy(&p.local_addr.su, local_addr.addr, local_addr.addrlen);
-  memcpy(&p.remote_addr.su, remote_addr.addr, remote_addr.addrlen);
+  memcpy(&p.local_addr.su, path.local.addr, path.local.addrlen);
+  memcpy(&p.remote_addr.su, path.remote.addr, path.remote.addrlen);
 
-  p.local_addr.len = local_addr.addrlen;
-  p.remote_addr.len = remote_addr.addrlen;
-  p.endpoint = &ep;
+  p.local_addr.len = path.local.addrlen;
+  p.remote_addr.len = path.remote.addrlen;
+  p.endpoint = static_cast<Endpoint *>(path.user_data);
   p.ecn = ecn;
   p.data = data;
   p.gso_size = gso_size;
@@ -1293,6 +1255,47 @@ int Handler::send_conn_close() {
                               /* ecn = */ 0, conn_closebuf_->data());
 }
 
+int Handler::send_conn_close(const Endpoint &ep, const Address &local_addr,
+                             const sockaddr *sa, socklen_t salen,
+                             const ngtcp2_pkt_info *pi,
+                             std::span<const uint8_t> data) {
+  assert(conn_closebuf_ && conn_closebuf_->size());
+
+  close_wait_.bytes_recv += data.size();
+  ++close_wait_.num_pkts_recv;
+
+  if (close_wait_.num_pkts_recv < close_wait_.next_pkts_recv ||
+      close_wait_.bytes_recv * 3 <
+        close_wait_.bytes_sent + conn_closebuf_->size()) {
+    return 0;
+  }
+
+  auto path = ngtcp2_path{
+    .local =
+      {
+        .addr = const_cast<sockaddr *>(&local_addr.su.sa),
+        .addrlen = local_addr.len,
+      },
+    .remote =
+      {
+        .addr = const_cast<sockaddr *>(sa),
+        .addrlen = salen,
+      },
+    .user_data = const_cast<Endpoint *>(&ep),
+  };
+
+  auto rv = server_->send_packet(ep, path.local, path.remote,
+                                 /* ecn = */ 0, conn_closebuf_->data());
+  if (rv != 0) {
+    return rv;
+  }
+
+  close_wait_.bytes_sent += conn_closebuf_->size();
+  close_wait_.next_pkts_recv *= 2;
+
+  return 0;
+}
+
 void Handler::update_timer() {
   auto expiry = ngtcp2_conn_get_expiry(conn_);
   auto now = util::timestamp();
@@ -1348,16 +1351,9 @@ int on_msg_complete(http_parser *htp) {
 } // namespace
 
 auto htp_settings = http_parser_settings{
-  on_msg_begin,    // on_message_begin
-  on_url_cb,       // on_url
-  nullptr,         // on_status
-  nullptr,         // on_header_field
-  nullptr,         // on_header_value
-  nullptr,         // on_headers_complete
-  nullptr,         // on_body
-  on_msg_complete, // on_message_complete
-  nullptr,         // on_chunk_header,
-  nullptr,         // on_chunk_complete
+  .on_message_begin = on_msg_begin,
+  .on_url = on_url_cb,
+  .on_message_complete = on_msg_complete,
 };
 
 int Handler::recv_stream_data(uint32_t flags, int64_t stream_id,
@@ -1367,7 +1363,7 @@ int Handler::recv_stream_data(uint32_t flags, int64_t stream_id,
   }
 
   auto it = streams_.find(stream_id);
-  assert(it != std::end(streams_));
+  assert(it != std::ranges::end(streams_));
   auto &stream = (*it).second;
 
   if (!stream->eos) {
@@ -1433,7 +1429,7 @@ int Handler::on_stream_close(int64_t stream_id, uint64_t app_error_code) {
   }
 
   auto it = streams_.find(stream_id);
-  assert(it != std::end(streams_));
+  assert(it != std::ranges::end(streams_));
   auto &stream = (*it).second;
 
   sendq_.erase(stream.get());
@@ -1506,7 +1502,7 @@ void Server::disconnect() {
   ev_signal_stop(loop_, &sigintev_);
 
   while (!handlers_.empty()) {
-    auto it = std::begin(handlers_);
+    auto it = std::ranges::begin(handlers_);
     auto &h = (*it).second;
 
     h->handle_error();
@@ -1526,14 +1522,14 @@ void Server::close() {
 namespace {
 int create_sock(Address &local_addr, const char *addr, const char *port,
                 int family) {
-  addrinfo hints{};
+  addrinfo hints{
+    .ai_flags = AI_PASSIVE,
+    .ai_family = family,
+    .ai_socktype = SOCK_DGRAM,
+  };
   addrinfo *res, *rp;
   int val = 1;
 
-  hints.ai_family = family;
-  hints.ai_socktype = SOCK_DGRAM;
-  hints.ai_flags = AI_PASSIVE;
-
   if (strcmp(addr, "*") == 0) {
     addr = nullptr;
   }
@@ -1722,24 +1718,26 @@ int Server::init(const char *addr, const char *port) {
   return 0;
 }
 
-int Server::on_read(Endpoint &ep) {
+int Server::on_read(const Endpoint &ep) {
   sockaddr_union su;
   std::array<uint8_t, 64_k> buf;
   size_t pktcnt = 0;
   ngtcp2_pkt_info pi;
 
-  iovec msg_iov;
-  msg_iov.iov_base = buf.data();
-  msg_iov.iov_len = buf.size();
-
-  msghdr msg{};
-  msg.msg_name = &su;
-  msg.msg_iov = &msg_iov;
-  msg.msg_iovlen = 1;
+  iovec msg_iov{
+    .iov_base = buf.data(),
+    .iov_len = buf.size(),
+  };
 
   uint8_t msg_ctrl[CMSG_SPACE(sizeof(int)) + CMSG_SPACE(sizeof(in6_pktinfo)) +
                    CMSG_SPACE(sizeof(int))];
-  msg.msg_control = msg_ctrl;
+
+  msghdr msg{
+    .msg_name = &su,
+    .msg_iov = &msg_iov,
+    .msg_iovlen = 1,
+    .msg_control = msg_ctrl,
+  };
 
   for (; pktcnt < 10;) {
     msg.msg_namelen = sizeof(su);
@@ -1753,8 +1751,8 @@ int Server::on_read(Endpoint &ep) {
       return 0;
     }
 
-    // Packets less than 22 bytes never be a valid QUIC packet.
-    if (nread < 22) {
+    // Packets less than 21 bytes never be a valid QUIC packet.
+    if (nread < 21) {
       ++pktcnt;
 
       continue;
@@ -1799,8 +1797,8 @@ int Server::on_read(Endpoint &ep) {
                   << std::dec << " " << datalen << " bytes" << std::endl;
       }
 
-      // Packets less than 22 bytes never be a valid QUIC packet.
-      if (datalen < 22) {
+      // Packets less than 21 bytes never be a valid QUIC packet.
+      if (datalen < 21) {
         break;
       }
 
@@ -1820,7 +1818,7 @@ int Server::on_read(Endpoint &ep) {
   return 0;
 }
 
-void Server::read_pkt(Endpoint &ep, const Address &local_addr,
+void Server::read_pkt(const Endpoint &ep, const Address &local_addr,
                       const sockaddr *sa, socklen_t salen,
                       const ngtcp2_pkt_info *pi,
                       std::span<const uint8_t> data) {
@@ -1844,7 +1842,7 @@ void Server::read_pkt(Endpoint &ep, const Address &local_addr,
   auto dcid_key = util::make_cid_key({vc.dcid, vc.dcidlen});
 
   auto handler_it = handlers_.find(dcid_key);
-  if (handler_it == std::end(handlers_)) {
+  if (handler_it == std::ranges::end(handlers_)) {
     ngtcp2_pkt_hd hd;
 
     if (auto rv = ngtcp2_accept(&hd, data.data(), data.size()); rv != 0) {
@@ -1853,7 +1851,7 @@ void Server::read_pkt(Endpoint &ep, const Address &local_addr,
                   << std::endl;
       }
 
-      if (!(data[0] & 0x80) && data.size() >= NGTCP2_SV_SCIDLEN + 22) {
+      if (!(data[0] & 0x80) && data.size() >= NGTCP2_SV_SCIDLEN + 21) {
         send_stateless_reset(data.size(), {vc.dcid, vc.dcidlen}, ep, local_addr,
                              sa, salen);
       }
@@ -1967,8 +1965,7 @@ void Server::read_pkt(Endpoint &ep, const Address &local_addr,
   auto h = (*handler_it).second;
   auto conn = h->conn();
   if (ngtcp2_conn_in_closing_period(conn)) {
-    // TODO do exponential backoff.
-    if (h->send_conn_close() != 0) {
+    if (h->send_conn_close(ep, local_addr, sa, salen, pi, data) != 0) {
       remove(h);
     }
     return;
@@ -2013,12 +2010,13 @@ uint32_t generate_reserved_version(const sockaddr *sa, socklen_t salen,
 int Server::send_version_negotiation(uint32_t version,
                                      std::span<const uint8_t> dcid,
                                      std::span<const uint8_t> scid,
-                                     Endpoint &ep, const Address &local_addr,
+                                     const Endpoint &ep,
+                                     const Address &local_addr,
                                      const sockaddr *sa, socklen_t salen) {
   Buffer buf{NGTCP2_MAX_UDP_PAYLOAD_SIZE};
   std::array<uint32_t, 1 + max_preferred_versionslen> sv;
 
-  auto p = std::begin(sv);
+  auto p = std::ranges::begin(sv);
 
   *p++ = generate_reserved_version(sa, salen, version);
 
@@ -2033,7 +2031,7 @@ int Server::send_version_negotiation(uint32_t version,
   auto nwrite = ngtcp2_pkt_write_version_negotiation(
     buf.wpos(), buf.left(), std::uniform_int_distribution<uint8_t>()(randgen),
     dcid.data(), dcid.size(), scid.data(), scid.size(), sv.data(),
-    p - std::begin(sv));
+    p - std::ranges::begin(sv));
   if (nwrite < 0) {
     std::cerr << "ngtcp2_pkt_write_version_negotiation: "
               << ngtcp2_strerror(nwrite) << std::endl;
@@ -2043,12 +2041,12 @@ int Server::send_version_negotiation(uint32_t version,
   buf.push(nwrite);
 
   ngtcp2_addr laddr{
-    const_cast<sockaddr *>(&local_addr.su.sa),
-    local_addr.len,
+    .addr = const_cast<sockaddr *>(&local_addr.su.sa),
+    .addrlen = local_addr.len,
   };
   ngtcp2_addr raddr{
-    const_cast<sockaddr *>(sa),
-    salen,
+    .addr = const_cast<sockaddr *>(sa),
+    .addrlen = salen,
   };
 
   if (send_packet(ep, laddr, raddr, /* ecn = */ 0, buf.data()) !=
@@ -2059,7 +2057,7 @@ int Server::send_version_negotiation(uint32_t version,
   return 0;
 }
 
-int Server::send_retry(const ngtcp2_pkt_hd *chd, Endpoint &ep,
+int Server::send_retry(const ngtcp2_pkt_hd *chd, const Endpoint &ep,
                        const Address &local_addr, const sockaddr *sa,
                        socklen_t salen, size_t max_pktlen) {
   std::array<char, NI_MAXHOST> host;
@@ -2116,12 +2114,12 @@ int Server::send_retry(const ngtcp2_pkt_hd *chd, Endpoint &ep,
   buf.push(nwrite);
 
   ngtcp2_addr laddr{
-    const_cast<sockaddr *>(&local_addr.su.sa),
-    local_addr.len,
+    .addr = const_cast<sockaddr *>(&local_addr.su.sa),
+    .addrlen = local_addr.len,
   };
   ngtcp2_addr raddr{
-    const_cast<sockaddr *>(sa),
-    salen,
+    .addr = const_cast<sockaddr *>(sa),
+    .addrlen = salen,
   };
 
   if (send_packet(ep, laddr, raddr, /* ecn = */ 0, buf.data()) !=
@@ -2133,7 +2131,7 @@ int Server::send_retry(const ngtcp2_pkt_hd *chd, Endpoint &ep,
 }
 
 int Server::send_stateless_connection_close(const ngtcp2_pkt_hd *chd,
-                                            Endpoint &ep,
+                                            const Endpoint &ep,
                                             const Address &local_addr,
                                             const sockaddr *sa,
                                             socklen_t salen) {
@@ -2150,12 +2148,12 @@ int Server::send_stateless_connection_close(const ngtcp2_pkt_hd *chd,
   buf.push(nwrite);
 
   ngtcp2_addr laddr{
-    const_cast<sockaddr *>(&local_addr.su.sa),
-    local_addr.len,
+    .addr = const_cast<sockaddr *>(&local_addr.su.sa),
+    .addrlen = local_addr.len,
   };
   ngtcp2_addr raddr{
-    const_cast<sockaddr *>(sa),
-    salen,
+    .addr = const_cast<sockaddr *>(sa),
+    .addrlen = salen,
   };
 
   if (send_packet(ep, laddr, raddr, /* ecn = */ 0, buf.data()) !=
@@ -2167,7 +2165,7 @@ int Server::send_stateless_connection_close(const ngtcp2_pkt_hd *chd,
 }
 
 int Server::send_stateless_reset(size_t pktlen, std::span<const uint8_t> dcid,
-                                 Endpoint &ep, const Address &local_addr,
+                                 const Endpoint &ep, const Address &local_addr,
                                  const sockaddr *sa, socklen_t salen) {
   if (stateless_reset_bucket_ == 0) {
     return 0;
@@ -2225,12 +2223,12 @@ int Server::send_stateless_reset(size_t pktlen, std::span<const uint8_t> dcid,
   buf.push(nwrite);
 
   ngtcp2_addr laddr{
-    const_cast<sockaddr *>(&local_addr.su.sa),
-    local_addr.len,
+    .addr = const_cast<sockaddr *>(&local_addr.su.sa),
+    .addrlen = local_addr.len,
   };
   ngtcp2_addr raddr{
-    const_cast<sockaddr *>(sa),
-    salen,
+    .addr = const_cast<sockaddr *>(sa),
+    .addrlen = salen,
   };
 
   if (send_packet(ep, laddr, raddr, /* ecn = */ 0, buf.data()) !=
@@ -2328,7 +2326,7 @@ int Server::verify_token(const ngtcp2_pkt_hd *hd, const sockaddr *sa,
   return 0;
 }
 
-int Server::send_packet(Endpoint &ep, const ngtcp2_addr &local_addr,
+int Server::send_packet(const Endpoint &ep, const ngtcp2_addr &local_addr,
                         const ngtcp2_addr &remote_addr, unsigned int ecn,
                         std::span<const uint8_t> data) {
   auto no_gso = false;
@@ -2339,7 +2337,8 @@ int Server::send_packet(Endpoint &ep, const ngtcp2_addr &local_addr,
 }
 
 std::pair<std::span<const uint8_t>, int>
-Server::send_packet(Endpoint &ep, bool &no_gso, const ngtcp2_addr &local_addr,
+Server::send_packet(const Endpoint &ep, bool &no_gso,
+                    const ngtcp2_addr &local_addr,
                     const ngtcp2_addr &remote_addr, unsigned int ecn,
                     std::span<const uint8_t> data, size_t gso_size) {
   assert(gso_size);
@@ -2356,7 +2355,7 @@ Server::send_packet(Endpoint &ep, bool &no_gso, const ngtcp2_addr &local_addr,
       auto len = std::min(gso_size, data.size());
 
       auto [_, rv] = send_packet(ep, no_gso, local_addr, remote_addr, ecn,
-                                 {std::begin(data), len}, len);
+                                 data.first(len), len);
       if (rv != 0) {
         return {data, rv};
       }
@@ -2367,23 +2366,22 @@ Server::send_packet(Endpoint &ep, bool &no_gso, const ngtcp2_addr &local_addr,
     return {{}, 0};
   }
 
-  iovec msg_iov;
-  msg_iov.iov_base = const_cast<uint8_t *>(data.data());
-  msg_iov.iov_len = data.size();
-
-  msghdr msg{};
-  msg.msg_name = const_cast<sockaddr *>(remote_addr.addr);
-  msg.msg_namelen = remote_addr.addrlen;
-  msg.msg_iov = &msg_iov;
-  msg.msg_iovlen = 1;
+  iovec msg_iov{
+    .iov_base = const_cast<uint8_t *>(data.data()),
+    .iov_len = data.size(),
+  };
 
   uint8_t msg_ctrl[CMSG_SPACE(sizeof(int)) + CMSG_SPACE(sizeof(uint16_t)) +
-                   CMSG_SPACE(sizeof(in6_pktinfo))];
+                   CMSG_SPACE(sizeof(in6_pktinfo))]{};
 
-  memset(msg_ctrl, 0, sizeof(msg_ctrl));
-
-  msg.msg_control = msg_ctrl;
-  msg.msg_controllen = sizeof(msg_ctrl);
+  msghdr msg{
+    .msg_name = const_cast<sockaddr *>(remote_addr.addr),
+    .msg_namelen = remote_addr.addrlen,
+    .msg_iov = &msg_iov,
+    .msg_iovlen = 1,
+    .msg_control = msg_ctrl,
+    .msg_controllen = sizeof(msg_ctrl),
+  };
 
   size_t controllen = 0;
 
@@ -2395,9 +2393,10 @@ Server::send_packet(Endpoint &ep, bool &no_gso, const ngtcp2_addr &local_addr,
     cm->cmsg_level = IPPROTO_IP;
     cm->cmsg_type = IP_PKTINFO;
     cm->cmsg_len = CMSG_LEN(sizeof(in_pktinfo));
-    in_pktinfo pktinfo{};
     auto addrin = reinterpret_cast<sockaddr_in *>(local_addr.addr);
-    pktinfo.ipi_spec_dst = addrin->sin_addr;
+    in_pktinfo pktinfo{
+      .ipi_spec_dst = addrin->sin_addr,
+    };
     memcpy(CMSG_DATA(cm), &pktinfo, sizeof(pktinfo));
 
     break;
@@ -2407,9 +2406,10 @@ Server::send_packet(Endpoint &ep, bool &no_gso, const ngtcp2_addr &local_addr,
     cm->cmsg_level = IPPROTO_IPV6;
     cm->cmsg_type = IPV6_PKTINFO;
     cm->cmsg_len = CMSG_LEN(sizeof(in6_pktinfo));
-    in6_pktinfo pktinfo{};
     auto addrin = reinterpret_cast<sockaddr_in6 *>(local_addr.addr);
-    pktinfo.ipi6_addr = addrin->sin6_addr;
+    in6_pktinfo pktinfo{
+      .ipi6_addr = addrin->sin6_addr,
+    };
     memcpy(CMSG_DATA(cm), &pktinfo, sizeof(pktinfo));
 
     break;
@@ -2500,12 +2500,10 @@ Server::send_packet(Endpoint &ep, bool &no_gso, const ngtcp2_addr &local_addr,
 }
 
 void Server::associate_cid(const ngtcp2_cid *cid, Handler *h) {
-  handlers_.emplace(util::make_cid_key(cid), h);
+  handlers_.emplace(*cid, h);
 }
 
-void Server::dissociate_cid(const ngtcp2_cid *cid) {
-  handlers_.erase(std::string{util::make_cid_key(cid)});
-}
+void Server::dissociate_cid(const ngtcp2_cid *cid) { handlers_.erase(*cid); }
 
 void Server::remove(const Handler *h) {
   auto conn = h->conn();
@@ -2531,47 +2529,56 @@ void Server::on_stateless_reset_regen() {
 }
 
 namespace {
-int parse_host_port(Address &dest, int af, const char *first,
-                    const char *last) {
-  if (std::distance(first, last) == 0) {
+int parse_host_port(Address &dest, int af, const std::string_view &host_port) {
+  if (host_port.empty()) {
     return -1;
   }
 
-  const char *host_begin, *host_end, *it;
+  auto first = std::ranges::begin(host_port);
+  auto last = std::ranges::end(host_port);
+
+  std::string_view hostv;
+
   if (*first == '[') {
-    host_begin = first + 1;
-    it = std::find(host_begin, last, ']');
+    ++first;
+
+    auto it = std::ranges::find(first, last, ']');
     if (it == last) {
       return -1;
     }
-    host_end = it;
-    ++it;
-    if (it == last || *it != ':') {
+
+    hostv = std::string_view{first, it};
+    first = it + 1;
+
+    if (first == last || *first != ':') {
       return -1;
     }
   } else {
-    host_begin = first;
-    it = std::find(host_begin, last, ':');
+    auto it = std::ranges::find(first, last, ':');
     if (it == last) {
       return -1;
     }
-    host_end = it;
+
+    hostv = std::string_view{first, it};
+    first = it;
   }
 
-  if (++it == last) {
+  if (++first == last) {
     return -1;
   }
-  auto svc_begin = it;
 
   std::array<char, NI_MAXHOST> host;
-  *std::copy(host_begin, host_end, std::begin(host)) = '\0';
+  *std::ranges::copy(hostv, std::ranges::begin(host)).out = '\0';
 
-  addrinfo hints{}, *res;
-  hints.ai_family = af;
-  hints.ai_socktype = SOCK_DGRAM;
+  addrinfo hints{
+    .ai_family = af,
+    .ai_socktype = SOCK_DGRAM,
+  };
+  addrinfo *res;
+  auto svc = first;
 
-  if (auto rv = getaddrinfo(host.data(), svc_begin, &hints, &res); rv != 0) {
-    std::cerr << "getaddrinfo: [" << host.data() << "]:" << svc_begin << ": "
+  if (auto rv = getaddrinfo(host.data(), svc, &hints, &res); rv != 0) {
+    std::cerr << "getaddrinfo: [" << host.data() << "]:" << svc << ": "
               << gai_strerror(rv) << std::endl;
     return -1;
   }
@@ -2600,32 +2607,33 @@ void print_usage() {
 
 namespace {
 void config_set_default(Config &config) {
-  config = Config{};
-  config.tx_loss_prob = 0.;
-  config.rx_loss_prob = 0.;
-  config.ciphers = util::crypto_default_ciphers();
-  config.groups = util::crypto_default_groups();
-  config.timeout = 30 * NGTCP2_SECONDS;
-  {
-    auto path = realpath(".", nullptr);
-    assert(path);
-    config.htdocs = path;
-    free(path);
-  }
-  config.mime_types_file = "/etc/mime.types"sv;
-  config.max_data = 1_m;
-  config.max_stream_data_bidi_remote = 256_k;
-  config.max_stream_data_uni = 256_k;
-  config.max_window = 6_m;
-  config.max_stream_window = 6_m;
-  config.max_streams_bidi = 100;
-  config.max_streams_uni = 3;
-  config.max_dyn_length = 20_m;
-  config.cc_algo = NGTCP2_CC_ALGO_CUBIC;
-  config.initial_rtt = NGTCP2_DEFAULT_INITIAL_RTT;
-  config.handshake_timeout = UINT64_MAX;
-  config.ack_thresh = 2;
-  config.initial_pkt_num = UINT32_MAX;
+  auto path = realpath(".", nullptr);
+  assert(path);
+  auto htdocs = std::string(path);
+  free(path);
+
+  config = Config{
+    .tx_loss_prob = 0.,
+    .rx_loss_prob = 0.,
+    .ciphers = util::crypto_default_ciphers(),
+    .groups = util::crypto_default_groups(),
+    .htdocs = std::move(htdocs),
+    .mime_types_file = "/etc/mime.types"sv,
+    .timeout = 30 * NGTCP2_SECONDS,
+    .max_data = 1_m,
+    .max_stream_data_bidi_remote = 256_k,
+    .max_stream_data_uni = 256_k,
+    .max_streams_bidi = 100,
+    .max_streams_uni = 3,
+    .max_window = 6_m,
+    .max_stream_window = 6_m,
+    .max_dyn_length = 20_m,
+    .cc_algo = NGTCP2_CC_ALGO_CUBIC,
+    .initial_rtt = NGTCP2_DEFAULT_INITIAL_RTT,
+    .handshake_timeout = UINT64_MAX,
+    .ack_thresh = 2,
+    .initial_pkt_num = UINT32_MAX,
+  };
 }
 } // namespace
 
@@ -2860,7 +2868,8 @@ int main(int argc, char **argv) {
       {"ack-thresh", required_argument, &flag, 30},
       {"initial-pkt-num", required_argument, &flag, 31},
       {"pmtud-probes", required_argument, &flag, 32},
-      {nullptr, 0, nullptr, 0}};
+      {},
+    };
 
     auto optidx = 0;
     auto c = getopt_long(argc, argv, "d:hqr:st:V", long_opts, &optidx);
@@ -2931,8 +2940,7 @@ int main(int argc, char **argv) {
         break;
       case 4:
         // --preferred-ipv4-addr
-        if (parse_host_port(config.preferred_ipv4_addr, AF_INET, optarg,
-                            optarg + strlen(optarg)) != 0) {
+        if (parse_host_port(config.preferred_ipv4_addr, AF_INET, optarg) != 0) {
           std::cerr << "preferred-ipv4-addr: could not use "
                     << std::quoted(optarg) << std::endl;
           exit(EXIT_FAILURE);
@@ -2940,8 +2948,8 @@ int main(int argc, char **argv) {
         break;
       case 5:
         // --preferred-ipv6-addr
-        if (parse_host_port(config.preferred_ipv6_addr, AF_INET6, optarg,
-                            optarg + strlen(optarg)) != 0) {
+        if (parse_host_port(config.preferred_ipv6_addr, AF_INET6, optarg) !=
+            0) {
           std::cerr << "preferred-ipv6-addr: could not use "
                     << std::quoted(optarg) << std::endl;
           exit(EXIT_FAILURE);
@@ -3113,7 +3121,7 @@ int main(int argc, char **argv) {
                     << max_preferred_versionslen << std::endl;
         }
         config.preferred_versions.resize(l.size());
-        auto it = std::begin(config.preferred_versions);
+        auto it = std::ranges::begin(config.preferred_versions);
         for (const auto &k : l) {
           if (k == "v1"sv) {
             *it++ = NGTCP2_PROTO_VER_V1;
@@ -3142,7 +3150,7 @@ int main(int argc, char **argv) {
         // --available-versions
         auto l = util::split_str(optarg);
         config.available_versions.resize(l.size());
-        auto it = std::begin(config.available_versions);
+        auto it = std::ranges::begin(config.available_versions);
         for (const auto &k : l) {
           if (k == "v1"sv) {
             *it++ = NGTCP2_PROTO_VER_V1;
